import json

import aiohttp
import logging

from airly.exceptions import AirlyError

_LOGGER = logging.getLogger(__name__)


class _EmptyFormat:
    def __format__(self, format_spec):
        return ''


class _RequestsHandler:
    """Internal class to create Airly requests"""

    def __init__(self, api_key, session: aiohttp.ClientSession, base_url,
                 language=None):
        self.headers = {
            'Accept': 'application/json',
            'Accept-Encoding': 'gzip',
            'apikey': api_key,
        }
        if language is not None:
            self.headers['Accept-Language'] = language
        self.base_url = base_url
        self.session = session

    async def get(self, request_path):
        url = self.base_url + request_path
        _LOGGER.debug("Sending request: " + url)
        async with self.session.get(url, headers=self.headers) as response:
            if response.status != 200:
                _LOGGER.warning("Invalid response from Airly API: %s",
                                response.status)
                raise AirlyError(response.status, await response.text())

            data = await response.json()
            _LOGGER.debug(json.dumps(data))
            return data

class _DictToObj(dict):
    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)



from airly._private import _EmptyFormat, _DictToObj


class Installation(_DictToObj):
    @property
    def location(self):
        return _DictToObj(self.get('location'))

    @property
    def address(self):
        return _DictToObj(self.get('address'))

    @property
    def sponsor(self):
        return _DictToObj(self.get('sponsor'))


class _InstallationsLoader:
    def __init__(self, requests_handler):
        self._rh = requests_handler

    _REQUEST_BY_ID_FORMAT = "installations/{:d}"
    _REQUEST_NEAREST_FORMAT = "installations/nearest?lat={:f}&lng={:f}" \
                              "&maxDistanceKM={:f}&maxResults={:d}"

    async def load_by_id(self, installation_id):
        data = await self._load(
            self._REQUEST_BY_ID_FORMAT.format(installation_id))
        return Installation(data)

    async def load_nearest(self, latitude, longitude,
                           max_distance_km=None, max_results=None):
        if max_distance_km is None:
            max_distance_km = _EmptyFormat()
        if max_results is None:
            max_results = _EmptyFormat()
        data = await self._load(self._REQUEST_NEAREST_FORMAT.format(
            latitude, longitude, max_distance_km, max_results))
        return [Installation(x) for x in data]

    def _load(self, request_path):
        return self._rh.get(request_path)



class AirlyError(Exception):
    """Raised when Airly APi request ended in error.

    Attributes:
        status_code - error code returned by Airly
        status - more detailed description
        """

    def __init__(self, status_code, status):
        self.status_code = status_code
        self.status = status



import re
from datetime import datetime
from enum import Enum
import logging
from airly import _private
from airly._private import _EmptyFormat, _DictToObj

_LOGGER = logging.getLogger(__name__)


class Measurement(_DictToObj):
    """Measurement for specific time period returned from Airly API"""

    # Following values are extracted from this API URL:
    # https://airapi.airly.eu/v2/meta/measurements?apikey=
    PM1 = "PM1"
    PM25 = "PM25"
    PM10 = "PM10"
    TEMPERATURE = "TEMPERATURE"
    HUMIDITY = "HUMIDITY"
    PRESSURE = "PRESSURE"
    NO2 = "NO2"
    O3 = "O3"

    MEASUREMENTS_TYPES = [
        PM1,
        PM25,
        PM10,
        TEMPERATURE,
        HUMIDITY,
        PRESSURE,
        NO2,
        O3,
    ]

    def _parse_list(self, key):
        result = []
        list_to_parse = self.get(key)
        if list_to_parse is not None:
            for e in list_to_parse:
                result.append(_DictToObj(e))
        return result

    def __init__(self, data: dict):
        super().__init__(data)

        # Parse date-times
        self.fromDateTime = self._parse_datetime(data.get('fromDateTime'))
        self.tillDateTime = self._parse_datetime(data.get('tillDateTime'))

        # Make popular measurements available directly,
        # i.e. instead of x.values['PM1'] make it accessible as x.pm1
        self.values = {
            x['name']: x['value'] for x in
            data['values']} if 'values' in data else {}
        for t in Measurement.MEASUREMENTS_TYPES:
            self.__setattr__(t.lower(),
                             self.values[t] if t in self.values else None)

        self.indexes = self._parse_list('indexes')
        self.standards = self._parse_list('standards')

    @staticmethod
    def _parse_datetime(x):
        if x is None:
            return None
        # full timestamp with milliseconds
        match = re.match(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+Z", x)
        if match:
            return datetime.strptime(x, "%Y-%m-%dT%H:%M:%S.%fZ")

        # timestamp missing milliseconds
        match = re.match(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", x)
        if match:
            return datetime.strptime(x, "%Y-%m-%dT%H:%M:%SZ")


class MeasurementsSession:
    """A class for polling for measurements from Airly API."""

    class Mode(Enum):
        INSTALLATION = 0
        NEAREST = 1
        POINT = 2

    _REQUEST_INSTALLATION_FORMAT = \
        "measurements/installation?installationId={:d}"
    _REQUEST_NEAREST_FORMAT = \
        "measurements/nearest?lat={:f}&lng={:f}&maxDistanceKM={:f}"
    _REQUEST_POINT_FORMAT = "measurements/point?lat={:f}&lng={:f}"

    def __init__(self,
                 requests_handler: _private,
                 mode: Mode,
                 **kwargs):
        """Initialize the connection to specific installation"""
        self.requests_handler = requests_handler
        if mode == self.Mode.INSTALLATION:
            self.request_path = self._REQUEST_INSTALLATION_FORMAT.format(
                kwargs['installation_id'])
        else:
            lat = kwargs['latitude']
            lng = kwargs['longitude']
            if mode == self.Mode.NEAREST:
                dist = kwargs.get('max_distance_km')
                self.request_path = self._REQUEST_NEAREST_FORMAT.format(
                    lat, lng, dist if dist is not None else _EmptyFormat())
            else:
                self.request_path = self._REQUEST_POINT_FORMAT.format(lat, lng)

        self.current = Measurement({})
        self.history = []
        self.forecast = []

    async def update(self):
        """Get measurements."""
        data = await self.requests_handler.get(self.request_path)
        self.current = Measurement(data['current'])
        self.history = [Measurement(x) for x in data['history']]
        self.forecast = [Measurement(x) for x in data['forecast']]



"""
Python wrapper for getting air quality data from Airly sensors.
"""
import logging

import aiohttp

from airly.installations import _InstallationsLoader
from airly.measurements import MeasurementsSession

_LOGGER = logging.getLogger(__name__)


class Airly:
    """Main class to perform Airly APi requests"""
    AIRLY_API_URL = "https://airapi.airly.eu/v2/"

    def __init__(self, api_key, session: aiohttp.ClientSession,
                 base_url=AIRLY_API_URL, language=None):
        from airly._private import _RequestsHandler
        self._rh = _RequestsHandler(api_key, session, base_url, language)
        self._installations = _InstallationsLoader(self._rh)

    def load_installation_by_id(self, installation_id):
        return self._installations.load_by_id(installation_id)

    def load_installation_nearest(self, latitude, longitude,
                                  max_distance_km=None, max_results=None):
        return self._installations.load_nearest(
            latitude, longitude,
            max_distance_km=max_distance_km, max_results=max_results)

    def create_measurements_session_installation(self, installation_id):
        return MeasurementsSession(
            self._rh, MeasurementsSession.Mode.INSTALLATION,
            installation_id=installation_id)

    def create_measurements_session_nearest(
            self, latitude, longitude, max_distance_km=None):
        return MeasurementsSession(
            self._rh,
            MeasurementsSession.Mode.NEAREST,
            latitude=latitude, longitude=longitude,
            max_distance_km=max_distance_km)

    def create_measurements_session_point(
            self, latitude, longitude):
        return MeasurementsSession(
            self._rh, MeasurementsSession.Mode.POINT,
            latitude=latitude, longitude=longitude)



import setuptools

REQUIRES = [
    'aiohttp>=3.5.4',
]

with open("README.rst") as fh:
    long_description = fh.read()

setuptools.setup(
    name='airly',
    version='1.0.0',
    description=
    'Python wrapper for getting air quality data from Airly sensors.',
    long_description=long_description,
    url='https://github.com/ak-ambi/python-airly',
    license='MIT',
    packages=['airly'],
    install_requires=REQUIRES,
    python_requires='>=3.6.0',
    author='Paweł Stankowski',
    author_email='ak_ambi@op.pl',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    test_suite='tests',
)
